@@ -1,12 +1,10 @@
 #include "pgroonga.h"
 
 #include "pgrn-column-name.h"
-#include "pgrn-compatible.h"
 #include "pgrn-convert.h"
 #include "pgrn-global.h"
 #include "pgrn-groonga.h"
 #include "pgrn-pg.h"
-#include "pgrn-row-level-security.h"
 #include "pgrn-wal.h"
 
 #include <catalog/catalog.h>
@@ -18,7 +16,6 @@ bool PGrnIsZlibAvailable;
 bool PGrnIsZstdAvailable;
 bool PGrnIsTemporaryIndexSearchAvailable;
 
-static grn_ctx *ctx = &PGrnContext;
 static struct PGrnBuffers *buffers = &PGrnBuffers;
 
 static bool
@@ -63,208 +60,25 @@ PGrnInitializeGroongaInformation(void)
 	GRN_OBJ_FIN(ctx, &grnIsSupported);
 }
 
-const char *
-PGrnInspect(grn_obj *object)
-{
-	grn_obj *buffer = &(PGrnBuffers.inspect);
-
-	GRN_BULK_REWIND(buffer);
-	{
-		grn_rc rc = ctx->rc;
-		grn_inspect(ctx, buffer, object);
-		ctx->rc = rc;
-	}
-	GRN_TEXT_PUTC(ctx, buffer, '\0');
-	return GRN_TEXT_VALUE(buffer);
-}
-
-const char *
-PGrnInspectName(grn_obj *object)
-{
-	static char name[GRN_TABLE_MAX_KEY_SIZE];
-	int nameSize;
-
-	{
-		grn_rc rc = ctx->rc;
-		nameSize = grn_obj_name(ctx, object, name, GRN_TABLE_MAX_KEY_SIZE);
-		name[nameSize] = '\0';
-		ctx->rc = rc;
-	}
-
-	return name;
-}
-
-int
-PGrnRCToPgErrorCode(grn_rc rc)
-{
-	int errorCode = ERRCODE_SYSTEM_ERROR;
-
-	/* TODO: Fill me. */
-	switch (rc)
-	{
-	case GRN_NO_SUCH_FILE_OR_DIRECTORY:
-		errorCode = ERRCODE_IO_ERROR;
-		break;
-	case GRN_INPUT_OUTPUT_ERROR:
-		errorCode = ERRCODE_IO_ERROR;
-		break;
-	case GRN_INVALID_ARGUMENT:
-		errorCode = ERRCODE_INVALID_PARAMETER_VALUE;
-		break;
-	case GRN_FUNCTION_NOT_IMPLEMENTED:
-		errorCode = ERRCODE_FEATURE_NOT_SUPPORTED;
-		break;
-	case GRN_NO_MEMORY_AVAILABLE:
-		errorCode = ERRCODE_OUT_OF_MEMORY;
-		break;
-	default:
-		break;
-	}
-
-	return errorCode;
-}
-
-bool
-PGrnCheck(const char *format, ...)
-{
-#define MESSAGE_SIZE 4096
-	va_list args;
-	char message[MESSAGE_SIZE];
-
-	if (ctx->rc == GRN_SUCCESS)
-		return true;
-
-	if (PGrnIsRLSEnabled)
-		PG_RE_THROW();
-
-	va_start(args, format);
-	grn_vsnprintf(message, MESSAGE_SIZE, format, args);
-	va_end(args);
-	ereport(ERROR,
-			(errcode(PGrnRCToPgErrorCode(ctx->rc)),
-			 errmsg("pgroonga: %s: %s", message, ctx->errbuf)));
-	return false;
-#undef MESSAGE_SIZE
-}
-
-bool
-PGrnCheckRC(grn_rc rc, const char *format, ...)
-{
-#define MESSAGE_SIZE 4096
-	va_list args;
-	char message[MESSAGE_SIZE];
-
-	if (rc == GRN_SUCCESS)
-		return true;
-
-	if (PGrnIsRLSEnabled)
-		PG_RE_THROW();
-
-	va_start(args, format);
-	grn_vsnprintf(message, MESSAGE_SIZE, format, args);
-	va_end(args);
-	ereport(ERROR,
-			(errcode(PGrnRCToPgErrorCode(rc)),
-			 errmsg("pgroonga: %s", message)));
-	return false;
-#undef MESSAGE_SIZE
-}
-
-bool
-PGrnCheckRCLevel(grn_rc rc, int errorLevel, const char *format, ...)
-{
-#define MESSAGE_SIZE 4096
-	va_list args;
-	char message[MESSAGE_SIZE];
-
-	if (rc == GRN_SUCCESS)
-		return true;
-
-	if (PGrnIsRLSEnabled)
-	{
-		if (errorLevel == ERROR)
-		{
-			PG_RE_THROW();
-		}
-		else
-		{
-			return false;
-		}
-	}
-
-	va_start(args, format);
-	grn_vsnprintf(message, MESSAGE_SIZE, format, args);
-	va_end(args);
-	ereport(errorLevel,
-			(errcode(PGrnRCToPgErrorCode(rc)),
-			 errmsg("pgroonga: %s", message)));
-	return false;
-#undef MESSAGE_SIZE
-}
-
 grn_obj *
 PGrnLookup(const char *name, int errorLevel)
 {
 	return PGrnLookupWithSize(name, strlen(name), errorLevel);
 }
 
-grn_obj *
-PGrnLookupWithSize(const char *name,
-				   size_t nameSize,
-				   int errorLevel)
-{
-	grn_obj *object;
-	object = grn_ctx_get(ctx, name, nameSize);
-	if (!object && errorLevel != PGRN_ERROR_LEVEL_IGNORE)
-	{
-		PGrnCheckRCLevel(GRN_INVALID_ARGUMENT,
-						 errorLevel,
-						 "object isn't found: <%.*s>",
-						 (int)nameSize, name);
-	}
-	return object;
-}
-
 grn_obj *
 PGrnLookupColumn(grn_obj *table, const char *name, int errorLevel)
 {
 	return PGrnLookupColumnWithSize(table, name, strlen(name), errorLevel);
 }
 
-grn_obj *
-PGrnLookupColumnWithSize(grn_obj *table,
-						 const char *name,
-						 size_t nameSize,
-						 int errorLevel)
-{
-	char columnName[GRN_TABLE_MAX_KEY_SIZE];
-	size_t columnNameSize;
-	grn_obj *column;
-
-	columnNameSize = PGrnColumnNameEncodeWithSize(name, nameSize, columnName);
-	column = grn_obj_column(ctx, table, columnName, columnNameSize);
-	if (!column)
-	{
-		char tableName[GRN_TABLE_MAX_KEY_SIZE];
-		int tableNameSize;
-
-		tableNameSize = grn_obj_name(ctx, table, tableName, sizeof(tableName));
-		PGrnCheckRCLevel(GRN_INVALID_ARGUMENT,
-						 errorLevel,
-						 "column isn't found: <%.*s>:<%.*s>",
-						 tableNameSize, tableName,
-						 (int) nameSize, name);
-	}
-
-	return column;
-}
-
 grn_obj *
 PGrnLookupSourcesTable(Relation index, int errorLevel)
 {
 	char name[GRN_TABLE_MAX_KEY_SIZE];
 
-	snprintf(name, sizeof(name),
+	snprintf(name,
+			 sizeof(name),
 			 PGrnSourcesTableNameFormat,
 			 PGRN_RELATION_GET_LOCATOR_NUMBER(index));
 	return PGrnLookup(name, errorLevel);
@@ -275,7 +89,8 @@ PGrnLookupSourcesCtidColumn(Relation index, int errorLevel)
 {
 	char name[GRN_TABLE_MAX_KEY_SIZE];
 
-	snprintf(name, sizeof(name),
+	snprintf(name,
+			 sizeof(name),
 			 PGrnSourcesTableNameFormat "." PGrnSourcesCtidColumnName,
 			 PGRN_RELATION_GET_LOCATOR_NUMBER(index));
 	return PGrnLookup(name, errorLevel);
@@ -286,7 +101,8 @@ PGrnLookupLexicon(Relation index, unsigned int nthAttribute, int errorLevel)
 {
 	char name[GRN_TABLE_MAX_KEY_SIZE];
 
-	snprintf(name, sizeof(name),
+	snprintf(name,
+			 sizeof(name),
 			 PGrnLexiconNameFormat,
 			 PGRN_RELATION_GET_LOCATOR_NUMBER(index),
 			 nthAttribute);
@@ -298,7 +114,8 @@ PGrnLookupIndexColumn(Relation index, unsigned int nthAttribute, int errorLevel)
 {
 	char name[GRN_TABLE_MAX_KEY_SIZE];
 
-	snprintf(name, sizeof(name),
+	snprintf(name,
+			 sizeof(name),
 			 PGrnLexiconNameFormat ".%s",
 			 PGRN_RELATION_GET_LOCATOR_NUMBER(index),
 			 nthAttribute,
@@ -314,9 +131,8 @@ PGrnFormatSourcesTableName(const char *indexName,
 	Oid fileNodeID;
 	indexID = PGrnPGIndexNameToID(indexName);
 	fileNodeID = PGrnPGIndexIDToFileNodeID(indexID);
-	snprintf(output, GRN_TABLE_MAX_KEY_SIZE,
-			 PGrnSourcesTableNameFormat,
-			 fileNodeID);
+	snprintf(
+		output, GRN_TABLE_MAX_KEY_SIZE, PGrnSourcesTableNameFormat, fileNodeID);
 }
 
 grn_obj *
@@ -353,49 +169,29 @@ PGrnCreateTableWithSize(Relation index,
 						grn_obj *normalizers,
 						grn_obj *tokenFilters)
 {
-	const char *path = NULL;
-	char pathBuffer[MAXPGPATH];
-	grn_obj	*table;
+	Oid indexTableSpaceID = InvalidOid;
+	grn_obj *table;
 
 	if (name)
 	{
 		flags |= GRN_OBJ_PERSISTENT;
-		if (index && PGRN_RELATION_GET_LOCATOR_SPACE(index) != MyDatabaseTableSpace)
-		{
-			char *databasePath;
-			char filePath[MAXPGPATH];
-
-			databasePath =
-				GetDatabasePath(MyDatabaseId, PGRN_RELATION_GET_LOCATOR_SPACE(index));
-			snprintf(filePath, sizeof(filePath),
-					 "%s.%.*s",
-					 PGrnDatabaseBasename,
-					 (int)nameSize,
-					 name);
-			join_path_components(pathBuffer,
-								 databasePath,
-								 filePath);
-			pfree(databasePath);
-
-			path = pathBuffer;
-		}
 	}
 
-	table = grn_table_create(ctx,
-							 name,
-							 nameSize,
-							 path,
-							 flags,
-							 type,
-							 NULL);
-	PGrnCheck("failed to create table: <%.*s>",
-			  (int)nameSize, name);
-	if (tokenizer)
-		grn_obj_set_info(ctx, table, GRN_INFO_DEFAULT_TOKENIZER, tokenizer);
-	if (normalizers)
-		grn_obj_set_info(ctx, table, GRN_INFO_NORMALIZERS, normalizers);
-	if (tokenFilters)
-		grn_obj_set_info(ctx, table, GRN_INFO_TOKEN_FILTERS, tokenFilters);
+	if (RelationIsValid(index))
+	{
+		indexTableSpaceID = PGRN_RELATION_GET_LOCATOR_SPACE(index);
+		if (indexTableSpaceID == MyDatabaseTableSpace)
+			indexTableSpaceID = InvalidOid;
+	}
+
+	table = PGrnCreateTableRawWithSize(indexTableSpaceID,
+									   name,
+									   nameSize,
+									   flags,
+									   type,
+									   tokenizer,
+									   normalizers,
+									   tokenFilters);
 
 	PGrnWALCreateTable(index,
 					   name,
@@ -410,10 +206,12 @@ PGrnCreateTableWithSize(Relation index,
 }
 
 grn_obj *
-PGrnCreateSimilarTemporaryLexicon(Relation index)
+PGrnCreateSimilarTemporaryLexicon(Relation index,
+								  const char *attributeName,
+								  size_t attributeNameSize,
+								  const char *tag)
 {
-	unsigned int i;
-	grn_obj *lexicon;
+	grn_obj *lexicon = NULL;
 	grn_table_flags flags = 0;
 	grn_obj *keyType = NULL;
 	grn_obj *tokenizer = NULL;
@@ -424,23 +222,59 @@ PGrnCreateSimilarTemporaryLexicon(Relation index)
 	grn_obj *tokenFiltersBuffer = &(buffers->tokenFilters);
 	grn_obj *temporaryLexicon = NULL;
 
-	for (i = 0; i < index->rd_att->natts; i++)
+	if (PGrnPGIsParentIndex(index))
 	{
-		lexicon = PGrnLookupLexicon(index, i, ERROR);
-		if (grn_type_id_is_text_family(ctx, lexicon->header.domain))
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("pgroonga: %s PGrnCreateSimilarTemporaryLexicon: "
+						"index must not a parent index: <%s%s%.*s>",
+						tag,
+						RelationGetRelationName(index),
+						attributeNameSize > 0 ? "." : "",
+						attributeNameSize > 0 ? (int) attributeNameSize : 0,
+						attributeName)));
+	}
+
+	if (attributeNameSize > 0)
+	{
+		int i = PGrnPGResolveAttributeIndex(
+			index, attributeName, attributeNameSize);
+		if (i != -1)
 		{
-			break;
+			lexicon = PGrnLookupLexicon(index, i, ERROR);
+			if (!grn_type_id_is_text_family(ctx, lexicon->header.domain))
+			{
+				grn_obj_unref(ctx, lexicon);
+				lexicon = NULL;
+			}
+		}
+	}
+	else
+	{
+		int i;
+		for (i = 0; i < index->rd_att->natts; i++)
+		{
+			lexicon = PGrnLookupLexicon(index, i, ERROR);
+			if (grn_type_id_is_text_family(ctx, lexicon->header.domain))
+			{
+				break;
+			}
+			grn_obj_unref(ctx, lexicon);
+			lexicon = NULL;
 		}
-		grn_obj_unref(ctx, lexicon);
-		lexicon = NULL;
 	}
 	if (!lexicon)
 	{
 		ereport(ERROR,
 				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("pgroonga: PGrnCreateSimilarTemporaryLexicon: "
-						"index doesn't have a lexicon for text: <%s>",
-						RelationGetRelationName(index))));
+				 errmsg("pgroonga: %s PGrnCreateSimilarTemporaryLexicon: "
+						"index doesn't have a lexicon for text: "
+						"<%s%s%.*s>",
+						tag,
+						RelationGetRelationName(index),
+						attributeNameSize > 0 ? "." : "",
+						attributeNameSize > 0 ? (int) attributeNameSize : 0,
+						attributeName)));
 	}
 
 	switch (lexicon->header.type)
@@ -462,29 +296,27 @@ PGrnCreateSimilarTemporaryLexicon(Relation index)
 
 	GRN_BULK_REWIND(tokenizerBuffer);
 	grn_table_get_default_tokenizer_string(ctx, lexicon, tokenizerBuffer);
-	if (GRN_TEXT_LEN(tokenizerBuffer) > 0) {
+	if (GRN_TEXT_LEN(tokenizerBuffer) > 0)
+	{
 		tokenizer = tokenizerBuffer;
 	}
 
 	GRN_BULK_REWIND(normalizersBuffer);
 	grn_table_get_normalizers_string(ctx, lexicon, normalizersBuffer);
-	if (GRN_TEXT_LEN(normalizersBuffer) > 0) {
+	if (GRN_TEXT_LEN(normalizersBuffer) > 0)
+	{
 		normalizers = normalizersBuffer;
 	}
 
 	GRN_BULK_REWIND(tokenFiltersBuffer);
 	grn_table_get_token_filters_string(ctx, lexicon, tokenFiltersBuffer);
-	if (GRN_TEXT_LEN(tokenFiltersBuffer) > 0) {
+	if (GRN_TEXT_LEN(tokenFiltersBuffer) > 0)
+	{
 		tokenFilters = tokenFiltersBuffer;
 	}
 
-	temporaryLexicon = PGrnCreateTable(index,
-									   NULL,
-									   flags,
-									   keyType,
-									   tokenizer,
-									   normalizers,
-									   tokenFilters);
+	temporaryLexicon = PGrnCreateTable(
+		index, NULL, flags, keyType, tokenizer, normalizers, tokenFilters);
 
 	grn_obj_unref(ctx, lexicon);
 	grn_obj_unref(ctx, keyType);
@@ -493,72 +325,41 @@ PGrnCreateSimilarTemporaryLexicon(Relation index)
 }
 
 grn_obj *
-PGrnCreateColumn(Relation	index,
-				 grn_obj	*table,
+PGrnCreateColumn(Relation index,
+				 grn_obj *table,
 				 const char *name,
 				 grn_column_flags flags,
-				 grn_obj	*type)
+				 grn_obj *type)
 {
-	return PGrnCreateColumnWithSize(index,
-									table,
-									name,
-									strlen(name),
-									flags,
-									type);
+	return PGrnCreateColumnWithSize(
+		index, table, name, strlen(name), flags, type);
 }
 
 grn_obj *
-PGrnCreateColumnWithSize(Relation	index,
-						 grn_obj	*table,
+PGrnCreateColumnWithSize(Relation index,
+						 grn_obj *table,
 						 const char *name,
-						 size_t		nameSize,
+						 size_t nameSize,
 						 grn_column_flags flags,
-						 grn_obj	*type)
+						 grn_obj *type)
 {
-	const char *path = NULL;
-	char pathBuffer[MAXPGPATH];
+	Oid indexTableSpaceID = InvalidOid;
 	grn_obj *column;
 
 	if (name)
 	{
 		flags |= GRN_OBJ_PERSISTENT;
-		if (index && PGRN_RELATION_GET_LOCATOR_SPACE(index) != MyDatabaseTableSpace)
-		{
-			char *databasePath;
-			char tableName[GRN_TABLE_MAX_KEY_SIZE];
-			int tableNameSize;
-			char filePath[MAXPGPATH];
-
-			databasePath =
-				GetDatabasePath(MyDatabaseId, PGRN_RELATION_GET_LOCATOR_SPACE(index));
-			tableNameSize = grn_obj_name(ctx,
-										 table,
-										 tableName,
-										 GRN_TABLE_MAX_KEY_SIZE);
-			snprintf(filePath, sizeof(filePath),
-					 "%s.%.*s.%.*s",
-					 PGrnDatabaseBasename,
-					 tableNameSize,
-					 tableName,
-					 (int)nameSize,
-					 name);
-			join_path_components(pathBuffer,
-								 databasePath,
-								 filePath);
-			pfree(databasePath);
-
-			path = pathBuffer;
-		}
 	}
-	column = grn_column_create(ctx,
-							   table,
-							   name,
-							   nameSize,
-							   path,
-							   flags,
-							   type);
-	PGrnCheck("failed to create column: <%.*s>",
-			  (int)nameSize, name);
+
+	if (RelationIsValid(index))
+	{
+		indexTableSpaceID = PGRN_RELATION_GET_LOCATOR_SPACE(index);
+		if (indexTableSpaceID == MyDatabaseTableSpace)
+			indexTableSpaceID = InvalidOid;
+	}
+
+	column = PGrnCreateColumnRawWithSize(
+		indexTableSpaceID, table, name, nameSize, flags, type);
 
 	PGrnWALCreateColumn(index, table, name, nameSize, flags, type);
 
@@ -566,17 +367,14 @@ PGrnCreateColumnWithSize(Relation	index,
 }
 
 void
-PGrnIndexColumnClearSources(Relation index,
-							grn_obj *indexColumn)
+PGrnIndexColumnClearSources(Relation index, grn_obj *indexColumn)
 {
 	GRN_BULK_REWIND(&(buffers->sourceIDs));
 	PGrnIndexColumnSetSourceIDs(index, indexColumn, &(buffers->sourceIDs));
 }
 
 void
-PGrnIndexColumnSetSource(Relation index,
-						 grn_obj *indexColumn,
-						 grn_obj *source)
+PGrnIndexColumnSetSource(Relation index, grn_obj *indexColumn, grn_obj *source)
 {
 	grn_id sourceID;
 
@@ -593,10 +391,9 @@ PGrnIndexColumnSetSourceIDs(Relation index,
 							grn_obj *indexColumn,
 							grn_obj *sourceIDs)
 {
-	grn_obj_set_info(ctx, indexColumn, GRN_INFO_SOURCE, sourceIDs);
-	PGrnCheck("failed to set sources: <%s>: <%s>",
-			  PGrnInspectName(indexColumn),
-			  PGrnInspect(sourceIDs));
+	if (!PGrnIndexColumnSetSourceIDsRaw(indexColumn, sourceIDs))
+		return;
+
 	PGrnWALSetSourceIDs(index, indexColumn, sourceIDs);
 }
 
@@ -609,16 +406,9 @@ PGrnRenameTable(Relation index, grn_obj *table, const char *newName)
 
 	nameSize = grn_obj_name(ctx, table, name, GRN_TABLE_MAX_KEY_SIZE);
 	newNameSize = strlen(newName);
-	grn_table_rename(ctx, table, newName, strlen(newName));
-	PGrnCheck("failed to rename table: <%s> -> <%s>",
-			  PGrnInspectName(table),
-			  newName);
+	PGrnRenameTableRawWithSize(table, newName, newNameSize);
 
-	PGrnWALRenameTable(index,
-					   name,
-					   nameSize,
-					   newName,
-					   newNameSize);
+	PGrnWALRenameTable(index, name, nameSize, newName, newNameSize);
 }
 
 void
@@ -628,15 +418,41 @@ PGrnRemoveObject(const char *name)
 }
 
 void
-PGrnRemoveObjectWithSize(const char *name,
-						 size_t nameSize)
+PGrnRemoveObjectWithSize(const char *name, size_t nameSize)
 {
 	grn_obj *object;
 
 	object = PGrnLookupWithSize(name, nameSize, ERROR);
 	grn_obj_remove(ctx, object);
-	PGrnCheck("failed to remove: <%.*s>",
-			  (int)nameSize, name);
+	PGrnCheck("failed to remove: <%.*s>", (int) nameSize, name);
+}
+
+void
+PGrnRemoveObjectForce(Relation index, const char *name)
+{
+	PGrnRemoveObjectForceWithSize(index, name, strlen(name));
+}
+
+void
+PGrnRemoveObjectForceWithSize(Relation index, const char *name, size_t nameSize)
+{
+	grn_obj *object;
+
+	object = PGrnLookupWithSize(name, nameSize, PGRN_ERROR_LEVEL_IGNORE);
+	if (object)
+	{
+		if (grn_obj_remove(ctx, object) != GRN_SUCCESS)
+		{
+			object = NULL;
+		}
+	}
+	if (!object)
+	{
+		grn_obj_remove_force(ctx, name, nameSize);
+	}
+	PGrnCheck("failed to remove: <%.*s>", (int) nameSize, name);
+
+	PGrnWALRemoveObject(index, name, nameSize);
 }
 
 void
@@ -658,40 +474,47 @@ PGrnFlushObject(grn_obj *object, bool recursive)
 		return;
 
 	nameSize = grn_obj_name(ctx, object, name, GRN_TABLE_MAX_KEY_SIZE);
-	PGrnCheck("failed to flush: <%.*s>",
-			  nameSize, name);
+	PGrnCheck("failed to flush: <%.*s>", nameSize, name);
 }
 
 void
-PGrnRemoveColumns(grn_obj *table)
+PGrnRemoveColumns(Relation index, grn_obj *table)
 {
 	grn_hash *columns;
 
-	columns = grn_hash_create(ctx,
-							  NULL,
-							  sizeof(grn_id),
-							  0,
-							  GRN_TABLE_HASH_KEY | GRN_HASH_TINY);
+	columns = grn_hash_create(
+		ctx, NULL, sizeof(grn_id), 0, GRN_TABLE_HASH_KEY | GRN_HASH_TINY);
 	if (!columns)
 	{
 		PGrnCheck("failed to create columns container for removing columns: "
 				  "<%s>",
 				  PGrnInspectName(table));
 	}
+	grn_table_columns(ctx, table, "", 0, (grn_obj *) columns);
+	PGrnCheck("failed to collect columns for removing columns: <%s>",
+			  PGrnInspectName(table));
 
-	GRN_HASH_EACH_BEGIN(ctx, columns, cursor, id) {
+	GRN_HASH_EACH_BEGIN(ctx, columns, cursor, id)
+	{
 		grn_id *columnID;
 		grn_obj *column;
+		char columnName[GRN_TABLE_MAX_KEY_SIZE];
+		int columnNameSize;
 
-		grn_hash_cursor_get_key(ctx, cursor, (void **)&columnID);
+		grn_hash_cursor_get_key(ctx, cursor, (void **) &columnID);
 		column = grn_ctx_at(ctx, *columnID);
 		if (!column)
 			continue;
 
+		columnNameSize =
+			grn_obj_name(ctx, column, columnName, GRN_TABLE_MAX_KEY_SIZE);
 		grn_obj_remove(ctx, column);
-		PGrnCheck("failed to remove column: <%s>",
-				  PGrnInspectName(column));
-	} GRN_HASH_EACH_END(ctx, cursor);
+		PGrnCheck(
+			"failed to remove column: <%.*s>", columnNameSize, columnName);
+
+		PGrnWALRemoveObject(index, columnName, columnNameSize);
+	}
+	GRN_HASH_EACH_END(ctx, cursor);
 
 	grn_hash_close(ctx, columns);
 }
@@ -732,7 +555,7 @@ PGrnPGTypeToGrnType(Oid pgTypeID, unsigned char *flags)
 		typeID = GRN_DB_LONG_TEXT;
 		break;
 	case VARCHAROID:
-		typeID = GRN_DB_SHORT_TEXT;	/* 4KB */
+		typeID = GRN_DB_SHORT_TEXT; /* 4KB */
 		break;
 #ifdef NOT_USED
 	case POINTOID:
@@ -832,3 +655,181 @@ PGrnGrnTypeToPGType(grn_id typeID)
 
 	return pgTypeID;
 }
+
+void
+PGrnExprAppendOp(grn_obj *expr,
+				 grn_operator op,
+				 int nArgs,
+				 const char *tag,
+				 const char *format,
+				 ...)
+{
+	grn_expr_append_op(ctx, expr, op, nArgs);
+	if (ctx->rc == GRN_SUCCESS)
+		return;
+
+	if (format)
+	{
+#define MESSAGE_SIZE 4096
+		va_list args;
+		char message[MESSAGE_SIZE];
+
+		va_start(args, format);
+		grn_vsnprintf(message, MESSAGE_SIZE, format, args);
+		va_end(args);
+
+		PGrnCheck("%s: failed to append operator: %s(%d): %s: %s",
+				  tag,
+				  grn_operator_to_string(op),
+				  nArgs,
+				  message,
+				  PGrnInspect(expr));
+#undef MESSAGE_SIZE
+	}
+	else
+	{
+		PGrnCheck("%s: failed to append operator: %s(%d): %s",
+				  tag,
+				  grn_operator_to_string(op),
+				  nArgs,
+				  PGrnInspect(expr));
+	}
+}
+
+void
+PGrnExprAppendConst(
+	grn_obj *expr, grn_obj *object, grn_operator op, int nArgs, const char *tag)
+{
+	grn_expr_append_const(ctx, expr, object, op, nArgs);
+	if (ctx->rc == GRN_SUCCESS)
+		return;
+
+	{
+#define BUFFER_SIZE 4096
+		char inspected[BUFFER_SIZE];
+		strncpy(inspected, PGrnInspect(object), BUFFER_SIZE - 1);
+		inspected[BUFFER_SIZE - 1] = '\0';
+#undef BUFFER_SIZE
+		PGrnCheck("%s: failed to %s(%d) a constant: %s: %s",
+				  tag,
+				  grn_operator_to_string(op),
+				  nArgs,
+				  inspected,
+				  PGrnInspect(expr));
+	}
+}
+
+void
+PGrnExprAppendConstString(grn_obj *expr,
+						  const char *string,
+						  unsigned int stringSize,
+						  grn_operator op,
+						  int nArgs,
+						  const char *tag)
+{
+	grn_expr_append_const_str(ctx, expr, string, stringSize, op, nArgs);
+	if (ctx->rc == GRN_SUCCESS)
+		return;
+
+	PGrnCheck("%s: failed to %s(%d) string: <%.*s>: %s",
+			  tag,
+			  grn_operator_to_string(op),
+			  nArgs,
+			  (int) stringSize,
+			  string,
+			  PGrnInspect(expr));
+}
+
+void
+PGrnExprAppendObject(grn_obj *expr,
+					 grn_obj *object,
+					 grn_operator op,
+					 int nArgs,
+					 const char *tag,
+					 const char *format,
+					 ...)
+{
+	grn_expr_append_obj(ctx, expr, object, op, nArgs);
+	if (ctx->rc == GRN_SUCCESS)
+		return;
+
+	{
+#define BUFFER_SIZE 4096
+		char inspected[BUFFER_SIZE];
+		strncpy(inspected, PGrnInspect(object), BUFFER_SIZE - 1);
+		inspected[BUFFER_SIZE - 1] = '\0';
+#undef BUFFER_SIZE
+
+		if (format)
+		{
+#define MESSAGE_SIZE 4096
+			va_list args;
+			char message[MESSAGE_SIZE];
+
+			va_start(args, format);
+			grn_vsnprintf(message, MESSAGE_SIZE, format, args);
+			va_end(args);
+
+			PGrnCheck("%s: failed to %s(%d) object: %s: %s: %s",
+					  tag,
+					  grn_operator_to_string(op),
+					  nArgs,
+					  inspected,
+					  message,
+					  PGrnInspect(expr));
+#undef MESSAGE_SIZE
+		}
+		else
+		{
+			PGrnCheck("%s: failed to %s(%d) object: %s: %s",
+					  tag,
+					  grn_operator_to_string(op),
+					  nArgs,
+					  inspected,
+					  PGrnInspect(expr));
+		}
+	}
+}
+
+void
+PGrnExprAppendConstInteger(grn_obj *expr,
+						   int value,
+						   grn_operator op,
+						   int nArgs,
+						   const char *tag,
+						   const char *format,
+						   ...)
+{
+	grn_expr_append_const_int(ctx, expr, value, op, nArgs);
+	if (ctx->rc == GRN_SUCCESS)
+		return;
+
+	if (format)
+	{
+#define MESSAGE_SIZE 4096
+		va_list args;
+		char message[MESSAGE_SIZE];
+
+		va_start(args, format);
+		grn_vsnprintf(message, MESSAGE_SIZE, format, args);
+		va_end(args);
+
+		PGrnCheck("%s: failed to %s(%d) integer: <%d>: %s: %s",
+				  tag,
+				  grn_operator_to_string(op),
+				  nArgs,
+				  value,
+				  message,
+				  PGrnInspect(expr));
+#undef MESSAGE_SIZE
+	}
+	else
+	{
+		PGrnCheck("%s: failed to %s(%d) integer: <%d>: %s",
+				  tag,
+				  grn_operator_to_string(op),
+				  nArgs,
+				  value,
+				  PGrnInspect(expr));
+	}
+}
